/*
 * drivers/storage/ata.c
 *
 * Copyright 2002 Jeff Garzik
 *
 * ide-probe.c copyrights:
 *  Copyright (C) 1994-1998  Linus Torvalds & authors (see below)
 *  Mostly written by Mark Lord <mlord@pobox.com>
 *                and Gadi Oxman <gadio@netvision.net.il>
 *                and Andre Hedrick <andre@linux-ide.org>
 *
 * linux/drivers/ide/ide_modes.h copyrights:
 *  Copyright (C) 1996  Linus Torvalds, Igor Abramov, and Mark Lord
 *
 * ide-taskfile.c copyrights:
 *  Copyright (C) 2000-2002     Michael Cornwell <cornwell@acm.org>
 *  Copyright (C) 2000-2002     Andre Hedrick <andre@linux-ide.org>
 *  Copyright (C) 2001-2002     Klaus Smolin
 *                                      IBM Storage Technology Division
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/jiffies.h>
#include <linux/cache.h>
#include <linux/kmod.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include "ata.h"
#include "stuff.h"

struct ata_ports ata_std_ports[ATA_STD_NCHAN] = {
	{
		lba_high:	0x1f5,
		lba_mid:	0x1f4,
		lba_low:	0x1f3,
		device:		0x1f6,
		nsector:	0x1f2,
		status:		0x1f7,
		command:	0x1f7,
		error:		0x1f1,
		features:	0x1f1,
		data:		0x1f0,
		altstatus:	0x3f6,
		devctl:		0x3f6,
	},

	{
		lba_high:	0x175,
		lba_mid:	0x174,
		lba_low:	0x173,
		device:		0x176,
		nsector:	0x172,
		status:		0x177,
		command:	0x177,
		error:		0x171,
		features:	0x171,
		data:		0x170,
		altstatus:	0x376,
		devctl:		0x376,
	}
};

struct ata_cmd ata_cmd_prep[] = {
	/* ATA_PREP_DEVDIAG */
	{
		flags:		ATA_CMD_FL_HAVE_DEVCTL,
		devctl:		ATA_NO_INTR,
		command:	ATA_CMD_EXEC_DEV_DIAG,
	},
	/* ATA_PREP_ID */
	{
		flags:		ATA_CMD_FL_HAVE_DEVCTL,
		devctl:		ATA_NO_INTR,
		command:	ATA_CMD_ID,
	},
	/* ATA_PREP_ID_PACKET */
	{
		flags:		ATA_CMD_FL_HAVE_DEVCTL,
		devctl:		ATA_NO_INTR,
		command:	ATA_CMD_ID_PACKET,
	},
};

/* magic numbers come from ATA-6 spec (sec 9.12). */
static u8 ata_signature[5] = { 0x01, 0x01, 0x00, 0x00, 0x00 };
static u8 atapi_signature[5] = { 0x01, 0x01, 0x14, 0xeb, 0x00 };

static LIST_HEAD(ata_classes);

#ifdef CONFIG_KMOD
static const struct {
	ata_device_t dev_type;
	const char *module_name;
} ata_module_list[] = {
	{ ata_type_disk, "ata-disk" },
};
#endif /* CONFIG_KMOD */

/**
 * ata_cmd_write_pio - write an ATA cmd to hardware using PIO
 * @ad: ATA device to which command is written
 * @cmd: command to send to hardware
 *
 * Write an ATA command to ioport ATA registers.  This is the lowest
 * level of command issuance.  It is assume that the caller is
 * handling state management, keeping track of the hardware, etc.
 *
 * Note, based on flagged_taskfile.  See comment above
 * struct ata_cmd.
 */
void ata_cmd_write_pio (struct ata_device *ad, struct ata_cmd *cmd)
{
	/* control, I want control! */
	/* handle any low-level command prep host driver needs to do
	 * right before sending command.  This used to be called
	 * "maskproc"
	 */
	if (ad->cmd_prep)
		ad->cmd_prep(ad, cmd);
	else {
		if (cmd->flags & ATA_CMD_FL_HAVE_DEVCTL)
			outb(cmd->devctl, ad->ports.devctl);
	}

	/* and then data... */
	if (cmd->flags & ATA_CMD_FL_HAVE_DATA) {
		u16 data = cmd->data + (cmd->hob_data << 8);
		outw(data, ad->ports.data);
	}

	/* send "hobfile" registers */
	if (cmd->flags & ATA_CMD_FL_HAVE_HOB_NSECTOR)
		outb(cmd->hob_nsector, ad->ports.nsector);
	if (cmd->flags & ATA_CMD_FL_HAVE_HOB_LBALOW)
		outb(cmd->hob_lba_low, ad->ports.lba_low);
	if (cmd->flags & ATA_CMD_FL_HAVE_HOB_LBAMID)
		outb(cmd->hob_lba_mid, ad->ports.lba_mid);
	if (cmd->flags & ATA_CMD_FL_HAVE_HOB_LBAHIGH)
		outb(cmd->hob_lba_high, ad->ports.lba_high);

	/* send regular registers */
	if (cmd->flags & ATA_CMD_FL_HAVE_FEATURE)
		outb(cmd->features, ad->ports.features);
	if (cmd->flags & ATA_CMD_FL_HAVE_NSECTOR)
		outb(cmd->nsector, ad->ports.nsector);
	if (cmd->flags & ATA_CMD_FL_HAVE_LBALOW)
		outb(cmd->lba_low, ad->ports.lba_low);
	if (cmd->flags & ATA_CMD_FL_HAVE_LBAMID)
		outb(cmd->lba_mid, ad->ports.lba_mid);
	if (cmd->flags & ATA_CMD_FL_HAVE_LBAHIGH)
		outb(cmd->lba_high, ad->ports.lba_high);

	/* device selection */
	outb(cmd->device, ad->ports.device);

	/* finally, issue the command */
	/* (ATA status and command registers are the same) */
	outb(cmd->command, ad->ports.command);
}

int ata_cmd_wait (struct ata_device *dev)
{
	unsigned int way_too_long = 1000;
	
	while (way_too_long-- > 0) {
		msleep(50);

		if (!(inb(dev->ports.status) & ATA_BUSY))
			return 0;
	}
	
	return -EIO;
}

static const struct ata_pio_timings ata_pio_timings[6] = {
	{ 70,	165,	600 },	/* PIO Mode 0 */
	{ 50,	125,	383 },	/* PIO Mode 1 */
	{ 30,	100,	240 },	/* PIO Mode 2 */
	{ 30,	80,	180 },	/* PIO Mode 3 with IORDY */
	{ 25,	70,	120 },	/* PIO Mode 4 with IORDY */
	{ 20,	50,	100 }	/* PIO Mode 5 with IORDY (nonstandard) */
};

/*
 * Black list. Some drives incorrectly report their maximal PIO mode,
 * at least in respect to CMD640. Here we keep info on some known drives.
 */
static const struct ata_pio_info {
	const char	*name;
	int		pio;
} ata_pio_blacklist [] = {
/*	{ "Conner Peripherals 1275MB - CFS1275A", 4 }, */
	{ "Conner Peripherals 540MB - CFS540A", 3 },

	{ "WDC AC2700",  3 },
	{ "WDC AC2540",  3 },
	{ "WDC AC2420",  3 },
	{ "WDC AC2340",  3 },
	{ "WDC AC2250",  0 },
	{ "WDC AC2200",  0 },
	{ "WDC AC21200", 4 },
	{ "WDC AC2120",  0 },
	{ "WDC AC2850",  3 },
	{ "WDC AC1270",  3 },
	{ "WDC AC1170",  1 },
	{ "WDC AC1210",  1 },
	{ "WDC AC280",   0 },
/*	{ "WDC AC21000", 4 }, */
	{ "WDC AC31000", 3 },
	{ "WDC AC31200", 3 },
/*	{ "WDC AC31600", 4 }, */

	{ "Maxtor 7131 AT", 1 },
	{ "Maxtor 7171 AT", 1 },
	{ "Maxtor 7213 AT", 1 },
	{ "Maxtor 7245 AT", 1 },
	{ "Maxtor 7345 AT", 1 },
	{ "Maxtor 7546 AT", 3 },
	{ "Maxtor 7540 AV", 3 },

	{ "SAMSUNG SHD-3121A", 1 },
	{ "SAMSUNG SHD-3122A", 1 },
	{ "SAMSUNG SHD-3172A", 1 },

/*	{ "ST51080A", 4 },
 *	{ "ST51270A", 4 },
 *	{ "ST31220A", 4 },
 *	{ "ST31640A", 4 },
 *	{ "ST32140A", 4 },
 *	{ "ST3780A",  4 },
 */
	{ "ST5660A",  3 },
	{ "ST3660A",  3 },
	{ "ST3630A",  3 },
	{ "ST3655A",  3 },
	{ "ST3391A",  3 },
	{ "ST3390A",  1 },
	{ "ST3600A",  1 },
	{ "ST3290A",  0 },
	{ "ST3144A",  0 },
	{ "ST3491A",  1 },	/* reports 3, should be 1 or 2 (depending on */	
				/* drive) according to Seagates FIND-ATA program */

	{ "QUANTUM ELS127A", 0 },
	{ "QUANTUM ELS170A", 0 },
	{ "QUANTUM LPS240A", 0 },
	{ "QUANTUM LPS210A", 3 },
	{ "QUANTUM LPS270A", 3 },
	{ "QUANTUM LPS365A", 3 },
	{ "QUANTUM LPS540A", 3 },
	{ "QUANTUM LIGHTNING 540A", 3 },
	{ "QUANTUM LIGHTNING 730A", 3 },

        { "QUANTUM FIREBALL_540", 3 }, /* Older Quantum Fireballs don't work */
        { "QUANTUM FIREBALL_640", 3 }, 
        { "QUANTUM FIREBALL_1080", 3 },
        { "QUANTUM FIREBALL_1280", 3 },
	{ NULL,	0 }
};

/*
 * This routine searches the ata_pio_blacklist for an entry
 * matching the start/whole of the supplied model name.
 *
 * Returns -1 if no match found.
 * Otherwise returns the recommended PIO mode from ata_pio_blacklist[].
 */
static int ata_scan_pio_blacklist (char *model)
{
	const struct ata_pio_info *p;

	for (p = ata_pio_blacklist; p->name != NULL; p++) {
		if (strncmp(p->name, model, strlen(p->name)) == 0)
			return p->pio;
	}
	return -1;
}

/*
 * This routine returns the recommended PIO settings for a given drive,
 * based on the drive->id information and the ata_pio_blacklist[].
 * This is used by most chipset support modules when "auto-tuning".
 */

/*
 * Drive PIO mode auto selection
 */
u8 ata_get_best_pio_mode (struct ata_device *dev, u8 mode_wanted,
			  u8 max_mode, struct ata_pio_data *d)
{
	int pio_mode;
	int cycle_time = 0;
	int use_iordy = 0;
	struct ata_id_page* id = &dev->id;
	int overridden  = 0;
	int blacklisted = 0;

	if (mode_wanted != 255) {
		pio_mode = mode_wanted;
	} else if ((pio_mode = ata_scan_pio_blacklist(id->model)) != -1) {
		overridden = 1;
		blacklisted = 1;
		use_iordy = (pio_mode > 2);
	} else {
		pio_mode = id->tPIO;
		if (pio_mode > 2) {	/* 2 is maximum allowed tPIO value */
			pio_mode = 2;
			overridden = 1;
		}
		if (id->field_valid & 2) {	  /* drive implements ATA2? */
			if (id->capability & 8) { /* drive supports use_iordy? */
				use_iordy = 1;
				cycle_time = id->eide_pio_iordy;
				if (id->eide_pio_modes & 7) {
					overridden = 0;
					if (id->eide_pio_modes & 4)
						pio_mode = 5;
					else if (id->eide_pio_modes & 2)
						pio_mode = 4;
					else
						pio_mode = 3;
				}
			} else {
				cycle_time = id->eide_pio;
			}
		}

		/*
		 * Conservative "downgrade" for all pre-ATA2 drives
		 */
		if (pio_mode && pio_mode < 4) {
			pio_mode--;
			overridden = 1;
			if (cycle_time && cycle_time < ata_pio_timings[pio_mode].cycle_time)
				cycle_time = 0; /* use standard timing */
		}
	}
	if (pio_mode > max_mode) {
		pio_mode = max_mode;
		cycle_time = 0;
	}
	if (d) {
		d->pio_mode = pio_mode;
		d->cycle_time = cycle_time ? cycle_time : ata_pio_timings[pio_mode].cycle_time;
		d->use_iordy = use_iordy;
		d->overridden = overridden;
		d->blacklisted = blacklisted;
	}
	return pio_mode;
}

static void ata_dev_new(struct ata_channel *chan, struct ata_device *fake_dev,
			unsigned int dev_num, unsigned int prep_idx)
{
	struct ata_device *dev;
	u16 config;
	u16 *id = (u16 *) &fake_dev->id;
	char *model;

	/* the id page is ready and waiting for us, read it ASAP */
	memset(id, 0, ATA_ID_PAGE_SIZE << 1);
	insw(fake_dev->ports.data, id, ATA_ID_PAGE_SIZE << 1);

	/* perform byte-swapping if necessary */	
#ifdef __BIG_ENDIAN
	{
		unsigned int i;
		
		for (i = 0; i < ATA_ID_PAGE_SIZE; i++)
			id[i] = le16_to_cpu(id[i]);
	}
#endif

	/* we do a lot of string stuff on 'model', so make it nice */
	fake_dev->id.model[sizeof(fake_dev->id.model) - 1] = 0;
	model = fake_dev->id.model;
	while (model[0] && (model[strlen(model) - 1] == ' '))
		model[strlen(model) - 1] = 0;

	/* allocate our new ATA device structure */
	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev)
		return;
	
	/* 'fake_dev' contains several already-set-up fields,
	 * so just copy to initialize
	 */
	memcpy(dev, fake_dev, sizeof(*dev));
	dev->device = dev_num;
	dev->chan = chan;
	dev->pdev = chan->host->sh.pdev;
	dev->lock = chan->lock;
	INIT_LIST_HEAD(&dev->list_node);
	INIT_LIST_HEAD(&dev->sd.list_node);
	dev->sd.bus_private = dev;
	
	/* add device to channel list */
	list_add_tail(&dev->list_node, &chan->dev_list);

	/*
	 * figure out what type of ATA[PI] device this is
	 */
	config = dev->id.config;
	dev->flags |= config & ATA_DEV_FL_REMOVEABLE;
	
	if (prep_idx != ATA_PREP_ID_PACKET) {
		dev->dev_type = ata_type_disk;
	} else {
		u8 type = (config >> 8) & 0x1f;

		switch (type) {
		case ata_type_floppy:
			if (strstr(dev->id.model, "CD-ROM"))
				type = ata_type_cdrom;
		case ata_type_cdrom:
		case ata_type_tape:
		case ata_type_optical:
			dev->dev_type = type;
			break;
		default:
			dev->dev_type = ata_type_unknown;
			break;
		}
	}
	
	/* set up PIO timings. attempt max PIO (it will dial down
	 * if necessary)
	 */
	dev->set_pio(dev, 5);
}

static void ata_dev_delete(struct ata_device *dev)
{
	/* XXX disable device, then free in ata_chan_shutdown? */
	list_del(&dev->list_node);
	kfree(dev);
}

static void ata_dev_probe (struct ata_channel *chan,
			   struct ata_device *fake_dev,
			   unsigned int dev_num,
			   unsigned int prep_idx)
{
	struct ata_cmd cmd;
	int rc;
	u8 status;
	
	/* set up and issue IDENTIFY [PACKET] DEVICE command */
	memcpy(&cmd, &ata_cmd_prep[prep_idx], sizeof(cmd));
	cmd.device = dev_num;

	ata_cmd_write_pio(fake_dev, &cmd);

	rc = ata_cmd_wait(fake_dev);
	if (rc)
		return;	
		
	/* check status. if command was aborted, it's likely an
	 * ATAPI device.  re-run probe.
	 */
	status = inb(fake_dev->ports.status);
	if (status & ATA_ERR) {
		u8 error;
		
		if (prep_idx == ATA_PREP_ID_PACKET)
			return;
			
		error = inb(fake_dev->ports.error);
		if (error & ATA_ABORTED_CMD) {
			u8 sig[5];
		
			sig[0] = inb(chan->ports.nsector);
			sig[1] = inb(chan->ports.lba_low);
			sig[2] = inb(chan->ports.lba_mid);
			sig[3] = inb(chan->ports.lba_high);
			sig[4] = inb(chan->ports.device);

			if (!memcmp(&sig, atapi_signature, 5))
				ata_dev_probe(chan, fake_dev, dev_num,
					      ATA_PREP_ID_PACKET);
		}
		
		return;
	}

	/* otherwise, we have a device, the ID page is ready to be read */
	ata_dev_new(chan, fake_dev, dev_num, prep_idx);
}

int ata_chan_init (struct ata_channel *chan)
{
	struct ata_device ad;
	int rc;
	u8 error;
	
	/* set up "fake" ATA device, to issue EXECUTE DEVICE DIAGNOSTIC
	 * command
	 */
	memset(&ad, 0, sizeof(ad));
	memcpy(&ad.ports, &chan->ports, sizeof(ad.ports));
	ad.cmd_prep = chan->host->info.cmd_prep;
	ad.set_pio = chan->host->info.set_pio;
	
	/* execute command */
	ata_cmd_write_pio(&ad, &ata_cmd_prep[ATA_PREP_DEVDIAG]);
	
	/* wait 2 msec, per spec */
	msleep(2);
	
	/* wait for host-idle */
	rc = ata_cmd_wait(&ad);
	if (rc)
		return rc;
		
	/* use error register diagnostic code to figure out
	 * what is on the channel.  unfortunately we don't
	 * really know whether device 0 or device 1 answered,
	 * so we can't take full advantage of the return codes
	 * and signature.
	 * magic numbers below come from EXEC DEV DIAG command docs
	 */
	error = inb(chan->ports.error);
	
	/* no devices on channel at all */
	if ((error == 0x80) || (error >= 0x82))
		return 0;
		
	/* device 0 is present, device 1 is not.  we can check
	 * the signature returned to see if it is an ATA or ATAPI device.
	 * fortunately this is often a common case, meaning our probe
	 * should occur more rapidly than previously in Linux.
	 */
	if (error == 0x81) {
		u8 sig[5];
		
		sig[0] = inb(chan->ports.nsector);
		sig[1] = inb(chan->ports.lba_low);
		sig[2] = inb(chan->ports.lba_mid);
		sig[3] = inb(chan->ports.lba_high);
		sig[4] = inb(chan->ports.device);

		msleep(50); /* cuz old ide driver did this */

		if (!memcmp(&sig, &ata_signature, 5))
			ata_dev_probe(chan, &ad, 0, ATA_PREP_ID);
		else if (!memcmp(&sig, atapi_signature, 5))
			ata_dev_probe(chan, &ad, 0, ATA_PREP_ID_PACKET);
			
		return 0;
	}

	msleep(50); /* cuz old ide driver did this */

	ata_dev_probe(chan, &ad, 0, ATA_PREP_ID);	

	msleep(50); /* cuz old ide driver did this */

	ata_dev_probe(chan, &ad, ATA_DEV1, ATA_PREP_ID);
	
	chan->flags |= ATA_CHAN_FL_ENABLED;
	return 0;
}

void ata_chan_shutdown (struct ata_channel *chan)
{
	while (!list_empty(&chan->dev_list)) {
		struct ata_device *dev;
		
		dev = list_entry(&chan->dev_list, struct ata_device, list_node);
		ata_dev_delete(dev);
	}	
	
	chan->flags &= ~ATA_CHAN_FL_ENABLED;
}

static int ata_dev_add (struct ata_device *dev)
{
	struct ata_class *cls;
	struct list_head *p;

#ifdef CONFIG_KMOD
	unsigned int i;
	
	for (i = 0; i < ARRAY_SIZE(ata_module_list); i++) {
		if (ata_module_list[i].dev_type == dev->dev_type) {
			request_module(ata_module_list[i].module_name);
			break;
		}
	}
#endif

	list_for_each(p, &ata_classes) {
		cls = list_entry(p, struct ata_class, list_node);
		if (cls->dev_type == dev->dev_type)
			return cls->add(dev);
	}
	
	return 0; /* XXX when 'null' class drvr exists, this changes */
}

static void ata_dev_remove (struct ata_device *dev)
{
	struct ata_class *cls;
	struct list_head *p;

	list_for_each(p, &ata_classes) {
		cls = list_entry(p, struct ata_class, list_node);
		if (cls->dev_type == dev->dev_type) {
			cls->remove(dev);
			return;
		}
	}
}

int ata_host_register (struct ata_host *host)
{
	unsigned int i;
	/* XXX register host with global storage system */
	
	for (i = 0; i < host->info.n_chan; i++) {
		struct list_head *p;
		
		list_for_each(p, &host->chan[i].dev_list) {
			struct ata_device *dev;
			
			dev = list_entry(p, struct ata_device, list_node);
			ata_dev_add(dev); /* XXX handle return val */
		}
	}

	return 0;
}

void ata_host_unregister (struct ata_host *host)
{
	unsigned int i;
	
	for (i = 0; i < host->info.n_chan; i++) {
		struct list_head *p;
		
		list_for_each(p, &host->chan[i].dev_list) {
			struct ata_device *dev;
			
			dev = list_entry(p, struct ata_device, list_node);
			ata_dev_remove(dev);
		}
	}

	/* XXX unregister host with global storage system */
}

static spinlock_t *alloc_aligned_lock(void **lock_base)
{
	void *mem = kmalloc(SMP_CACHE_BYTES * 2, GFP_KERNEL);
	unsigned long mem_val = (unsigned long) mem;
	if (!mem)
		return NULL;
	*lock_base = mem;
	mem_val += SMP_CACHE_BYTES;
	mem_val &= ~(SMP_CACHE_BYTES - 1);
	return (spinlock_t *) (void *) mem_val;
}

struct ata_host *ata_host_alloc (struct ata_host_info *host_info)
{
	struct ata_host *host;
	unsigned int alloc_size, i, n_chan, lock_strategy;
	void *lock_base = NULL;
	spinlock_t *lock = NULL;
	
	lock_strategy = host_info->lock_strategy;
	n_chan = host_info->n_chan;

	alloc_size = sizeof(struct ata_host) +
		     (n_chan * sizeof(struct ata_channel));
	
	host = kmalloc(alloc_size, GFP_KERNEL);
	if (!host)
		return NULL;
		
	memset(host, 0, alloc_size);
	memcpy(&host->info, host_info, sizeof(*host_info));

	if (lock_strategy == ATA_LOCK_HOST) {
		lock = alloc_aligned_lock(&lock_base);
		if (!lock)
			goto err_out_host;
	}
	
	for (i = 0; i < n_chan; i++) {
		host->chan[i].id = i;
		host->chan[i].host = host;
		INIT_LIST_HEAD(&host->chan[i].dev_list);
		
		if (lock_strategy == ATA_LOCK_CHANNEL) {
			lock = alloc_aligned_lock(&lock_base);
			if (!lock)
				goto err_out_chan;
			host->chan[i].lock = lock;
			host->chan[i].lock_base = lock_base;
		} else
			host->chan[i].lock = lock;
	}
	
	return host;

err_out_chan:
	for (i = 0; i < n_chan; i++)
		if (host->chan[i].lock_base)
			kfree(host->chan[i].lock_base);
err_out_host:
	kfree(host);
	return NULL;
}

void ata_host_free(struct ata_host *host)
{
	unsigned int i;
	
	if (host->info.lock_strategy == ATA_LOCK_HOST)
		kfree(host->chan[0].lock_base);
	else
		for (i = 0; i < host->info.n_chan; i++)
			kfree(host->chan[i].lock_base);
	
	kfree(host);
}

int ata_class_register(struct ata_class *cls)
{
	list_add_tail(&cls->list_node, &ata_classes);
	return 0;
}

void ata_class_unregister(struct ata_class *cls)
{
	list_del(&cls->list_node);
}

void ata_pci_legacy_fixup (struct pci_dev *pdev)
{
	u16 pci_command;
	unsigned long addr[4] = { 0x1f0, 0x3f6, 0x170, 0x376 };
	unsigned long len[4] = { 8, 1, 8, 1 };
	unsigned int i;
	int pci_region[4] = {
		PCI_BASE_ADDRESS_0,
		PCI_BASE_ADDRESS_1,
		PCI_BASE_ADDRESS_2,
		PCI_BASE_ADDRESS_3,
	};
	
	/* see if fixup is necessary */
	if ((pci_resource_start(pdev, 0)) &&
	    (pci_resource_start(pdev, 1)) &&
	    (pci_resource_start(pdev, 2)) &&
	    (pci_resource_start(pdev, 3)))
		return;
	
	/* disable IO decoding */
	pci_read_config_word(pdev, PCI_COMMAND, &pci_command);
	if (pci_command & PCI_COMMAND_IO)
		pci_write_config_word(pdev, PCI_COMMAND,
				      pci_command & ~PCI_COMMAND_IO);
	
	/* assign addresses as per array values above */
	for (i = 0; i < 4; i++) {
		pdev->resource[i].start = addr[i];
		pdev->resource[i].end = addr[i] + len[i] - 1;
		pdev->resource[i].flags = IORESOURCE_IO;
		pci_write_config_dword(pdev, pci_region[i],
				       pci_resource_start(pdev, i));
	}
	
	/* enable IO decoding */
	if (pci_command & PCI_COMMAND_IO)
		pci_write_config_word(pdev, PCI_COMMAND, pci_command);
}

EXPORT_SYMBOL(ata_cmd_prep);
EXPORT_SYMBOL(ata_cmd_wait);
EXPORT_SYMBOL(ata_cmd_write_pio);
EXPORT_SYMBOL(ata_std_ports);
EXPORT_SYMBOL(ata_chan_init);
EXPORT_SYMBOL(ata_chan_shutdown);
EXPORT_SYMBOL(ata_host_register);
EXPORT_SYMBOL(ata_host_unregister);
EXPORT_SYMBOL(ata_host_alloc);
EXPORT_SYMBOL(ata_pci_legacy_fixup);
EXPORT_SYMBOL(ata_get_best_pio_mode);
EXPORT_SYMBOL(ata_class_register);
EXPORT_SYMBOL(ata_class_unregister);

MODULE_AUTHOR("Jeff Garzik");
MODULE_DESCRIPTION("ATA core driver");
MODULE_LICENSE("GPL");
